package org.jastadd.relast.plugin;

import org.gradle.api.DefaultTask;
import org.gradle.api.Project;
import org.gradle.api.file.FileCollection;
import org.gradle.api.plugins.JavaPlugin;
import org.gradle.api.tasks.SourceSet;
import org.gradle.api.tasks.SourceSetContainer;
import org.gradle.api.tasks.TaskAction;

import java.nio.file.Paths;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

/**
 * RelAst Test Task
 *
 * @author rschoene - Initial contribution
 */
@SuppressWarnings({"unused", "WeakerAccess"})
public class RelastTest extends DefaultTask {
  // configuration from plugin
  private String compilerLocation;
  // general options
  private boolean verbose = false;
  // pre-process options
  private List<String> relastFiles = new ArrayList<>();
  private boolean useJastAddNames;
  private boolean resolverHelper;
  private boolean writeToFile = true;
  private String grammarName;
  private String listClass;
  private String jastAddList;
  private String serializer;

  // compile options
  private boolean runJastAdd = true;
  private String outputDir = "src/test/java-gen/";
  private String packageName;
  private final List<String> moreInputFiles = new ArrayList<>();

  public void setCompilerLocation(String compilerLocation) {
    this.compilerLocation = compilerLocation;
  }

  public boolean isVerbose() {
    return verbose;
  }

  public void setVerbose(boolean verbose) {
    this.verbose = verbose;
  }

  // pre-process options
  public List<String> getRelastFiles() {
    return relastFiles;
  }

  public void relastFiles(String relastFile) {
    this.relastFiles.add(relastFile);
  }

  public void relastFiles(String[] relastFilesArray) {
    this.relastFiles = Arrays.asList(relastFilesArray);
  }

  public boolean isUseJastAddNames() {
    return useJastAddNames;
  }

  public void setUseJastAddNames(boolean useJastAddNames) {
    this.useJastAddNames = useJastAddNames;
  }

  public boolean isResolverHelper() {
    return resolverHelper;
  }

  public void setResolverHelper(boolean resolverHelper) {
    this.resolverHelper = resolverHelper;
  }

  public boolean isWriteToFile() {
    return writeToFile;
  }

  public void setWriteToFile(boolean writeToFile) {
    this.writeToFile = writeToFile;
  }

  public String getGrammarName() {
    return grammarName;
  }

  public void setGrammarName(String grammarName) {
    this.grammarName = grammarName;
  }

  public String getListClass() {
    return listClass;
  }

  public void setListClass(String listClass) {
    this.listClass = listClass;
  }

  public String getJastAddList() {
    return jastAddList;
  }

  public void setJastAddList(String jastAddList) {
    this.jastAddList = jastAddList;
  }

  public String getSerializer() {
    return serializer;
  }

  public void setSerializer(String serializer) {
    this.serializer = serializer;
  }

  // compile options
  public boolean isRunJastAdd() {
    return runJastAdd;
  }

  public void setRunJastAdd(boolean runJastAdd) {
    this.runJastAdd = runJastAdd;
  }

  public String getOutputDir() {
    return outputDir;
  }

  public void setOutputDir(String outputDir) {
    this.outputDir = outputDir;
  }

  public String getPackageName() {
    return packageName;
  }

  public void setPackageName(String packageName) {
    this.packageName = packageName;
  }

  public List<String> getMoreInputFiles() {
    return moreInputFiles;
  }

  public void moreInputFiles(String f) {
    this.moreInputFiles.add(f);
  }

  public void moreInputFiles(String[] fileArray) {
    this.moreInputFiles.addAll(Arrays.asList(fileArray));
  }

  private boolean isSet(String option) {
    return option != null && !option.isEmpty();
  }

  private String[] genSuffixes = {".ast", ".jadd", "RefResolver.jadd", "ResolverStubs.jrag", "Serializer.jadd"};

  @TaskAction
  void runTest() {
    setGroup("verification");
    setDescription("Runs a relast test");
    Project project = getProject();
    if (isVerbose()) {
      System.out.println("Running relast test");
      System.out.println("relast files: " + getRelastFiles());
      System.out.println("Deleting files");
    }
    // first, delete generated files
    List<String> genFiles = new ArrayList<>();
    for (String suffix : genSuffixes) {
      genFiles.add(getGrammarName() + suffix);
    }
    if (isVerbose()) {
      System.out.println("gen files: " + genFiles);
    }
    project.delete(deleteSpec -> {
      deleteSpec.delete(genFiles);
      if (isSet(getPackageName())) {
        deleteSpec.delete(Paths.get(getOutputDir(), getPackageName()));
      }
    });
    if (isVerbose()) {
      System.out.println("Pre processing, running relast");
    }
    // then, run relast pre processing
    project.getPlugins().withType(JavaPlugin.class, javaPlugin -> {
      SourceSetContainer sourceSets = (SourceSetContainer) project.getProperties().get("sourceSets");
      FileCollection runtimeClasspath = sourceSets.getByName(SourceSet.MAIN_SOURCE_SET_NAME).getRuntimeClasspath();
      project.javaexec(javaExecSpec -> {
        List<Object> args = new ArrayList<>();
        javaExecSpec.setClasspath(runtimeClasspath);
        if (compilerLocation != null) {
          javaExecSpec.setMain("-jar");
          args.add(compilerLocation);
        } else {
          javaExecSpec.setMain("org.jastadd.relast.compiler.Compiler");
        }
        args.addAll(getRelastFiles());
        args.add("--quiet");
        if (isWriteToFile()) {
          args.add("--file");
        }
        if (isUseJastAddNames()) {
          args.add("--useJastAddNames");
        }
        if (isResolverHelper()) {
          args.add("--resolverHelper");
        }
        if (isSet(getJastAddList())) {
          args.add("--jastAddList=" + getJastAddList());
        }
        if (isSet(getListClass())) {
          args.add("--listClass=" + getListClass());
        }
        if (isSet(getSerializer())) {
          args.add("--serializer=" + getSerializer());
        }
        args.add("--grammarName=" + getGrammarName());
        if (isVerbose()) {
          System.out.println("Start relast with args: " + args);
        }
        javaExecSpec.args(args);
      });
    });
    if (isRunJastAdd()) {
      if (isVerbose()) {
        System.out.println("Compile with JastAdd");
      }
      // check which files were actually generated
      genFiles.removeIf(s -> !Paths.get(s).toFile().exists());
      // finally, compile generated files
      project.getPlugins().withType(JavaPlugin.class, javaPlugin -> {
        SourceSetContainer sourceSets = (SourceSetContainer) project.getProperties().get("sourceSets");
        FileCollection runtimeClasspath = sourceSets.getByName(SourceSet.MAIN_SOURCE_SET_NAME).getRuntimeClasspath();
        project.javaexec(javaExecSpec -> {
          javaExecSpec.setClasspath(runtimeClasspath);
          javaExecSpec.setMain("org.jastadd.JastAdd");
          List<Object> args = new ArrayList<>();
          args.add("--o=" + getOutputDir());
          args.add("--package=" + getPackageName());
          if (isSet(getJastAddList())) {
            args.add("--List=" + getJastAddList());
          }
          args.addAll(genFiles);
          args.addAll(getMoreInputFiles());
          if (isVerbose()) {
            System.out.println("Start JastAdd with args: " + args);
          }
          javaExecSpec.args(args);
        });
      });
    }
  }

}
