package de.tudresden.inf.st.ros2rag.example;

import de.tudresden.inf.st.ros2rag.ast.*;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import panda.Linkstate;

import java.io.IOException;
import java.util.concurrent.TimeUnit;

/**
 * Testing Ros2Rag without generating something.
 *
 * @author rschoene - Initial contribution
 */
public class Main {

  private static final String TOPIC_JOINT1 = "robot/joint1";
  private static final String TOPIC_CONFIG = "robot/config";

  private static final Logger logger = LogManager.getLogger(Main.class);

  public static void main(String[] args) throws IOException, InterruptedException {
    Model model = new Model();
    model.MqttSetHost("localhost");

    ZoneModel zoneModel = new ZoneModel();
    zoneModel.setSize(makePosition(1, 1, 1));

    IntPosition myPosition = makePosition(0, 0, 0);
    Coordinate myCoordinate = new Coordinate(myPosition);
    Coordinate leftPosition = new Coordinate(makePosition(-1, 0, 0));
    Coordinate rightPosition = new Coordinate(makePosition(1, 0, 0));

    Zone safetyZone = new Zone();
    safetyZone.addCoordinate(myCoordinate);
    safetyZone.addCoordinate(leftPosition);
    safetyZone.addCoordinate(rightPosition);
    zoneModel.addSafetyZone(safetyZone);
    model.setZoneModel(zoneModel);

    RobotArm robotArm = new RobotArm();
    robotArm.set_AttributeTestSource(1);  // set initial value, no trigger

    Joint joint1 = new Joint();
    joint1.setName("joint1");
    joint1.setCurrentPosition(myPosition);

    EndEffector endEffector = new EndEffector();
    endEffector.setName("gripper");
    endEffector.setCurrentPosition(makePosition(2, 2, 3));

    robotArm.addJoint(joint1);
    robotArm.setEndEffector(endEffector);
    model.setRobotArm(robotArm);

    // add dependencies
    robotArm.addDependency1(joint1);
    robotArm.addDependency1(endEffector);
    robotArm.addDependency2(robotArm);

    model.MqttWaitUntilReady(2, TimeUnit.SECONDS);

    joint1.connectCurrentPosition(TOPIC_JOINT1);
    robotArm.connect_AppropriateSpeed(TOPIC_CONFIG, true);

    startDelayedSendOne();

    logStatus("BEFORE", robotArm, joint1);
//    logger.info("Now invoke ./send_one.sh");

//    Thread.sleep(3000);
//    robotArm.setAttributeTestSource(1);
//    Thread.sleep(3000);
//    robotArm.setAttributeTestSource(5);
    Thread.sleep(6000);

    logStatus("AFTER", robotArm, joint1);

    logger.info("Now changing AttributeTestSource - this should not emit any new mqtt message");
    robotArm.setAttributeTestSource(1);
    Thread.sleep(1000);
    robotArm.setAttributeTestSource(5);

    model.MqttCloseConnections();
  }

  private static void logStatus(String prefix, RobotArm robotArm, Joint joint1) {
    logger.info("{} joint1.getCurrentPosition() = {}, " +
            "robotArm.isInSafetyZone = {}, " +
            "robotArm.get_AppropriateSpeed = {}",
        prefix,
        joint1.getCurrentPosition(),
        robotArm.isInSafetyZone(),
        robotArm.get_AppropriateSpeed());
  }

  private static void startDelayedSendOne() {
    final int millis = 1500;
    new Thread(() -> {
      try {
        Linkstate.PandaLinkState pls = Linkstate.PandaLinkState.newBuilder()
            .setName("Joint1")
            .setPos(Linkstate.PandaLinkState.Position.newBuilder()
                .setPositionX(0.5f)
                .setPositionY(0.5f)
                .setPositionZ(0.5f)
                .build())
            .build();
        byte[] bytes = pls.toByteArray();
        MqttUpdater mqttUpdater = new MqttUpdater("sender").setHost("localhost", 1883);

        // wait until connected, but in total millis
        long before = System.currentTimeMillis();
        mqttUpdater.waitUntilReady(millis, TimeUnit.MILLISECONDS);
        long alreadyWaited = System.currentTimeMillis() - before;
        Thread.sleep((long) millis - alreadyWaited);

        mqttUpdater.publish(TOPIC_JOINT1, bytes);

        Thread.sleep(1000);
        pls = Linkstate.PandaLinkState.newBuilder()
            .setName("Joint1")
            .setPos(Linkstate.PandaLinkState.Position.newBuilder()
                .setPositionX(2.5f)
                .setPositionY(2.5f)
                .setPositionZ(2.5f)
                .build())
            .build();
        bytes = pls.toByteArray();
        mqttUpdater.publish(TOPIC_JOINT1, bytes);

        mqttUpdater.close();
      } catch (IOException | InterruptedException e) {
        e.printStackTrace();
      }
    }).start();
  }

  private static IntPosition makePosition(int x, int y, int z) {
    return IntPosition.of(x, y, z);
  }
}
